#import "ES2Renderer.h"
#import "Shaders.h"
#include "matrix.h"
#include <mach/mach_time.h>

// ---- profiler bits

typedef   signed long long Prof_Int64;
mach_timebase_info_data_t s_TimeInfo;
static float MachToMillisecondsDelta (Prof_Int64 delta)
{
	// Convert to nanoseconds
	delta *= s_TimeInfo.numer;
	delta /= s_TimeInfo.denom;
	float result = (float)delta / 1000000.0F;
	return result;
}
static Prof_Int64 s_TimeAcc;


// ---- rendering

const float kLightR = 1.0f;
const float kLightG = 0.95f;
const float kLightB = 0.8f;
const float kSpec = 64.0f;
const int kLUTX = 32;
const int kLUTY = 128;

enum {
	kUniformMatrixMVP,
	kUniformMatrixWorld2Object,
	kUniformVecLightDir,
	kUniformVecCamPos,
	kUniformVecLightColor,
	kUniformVecMatColor,
	kUniformFloatSpec,
	kUniformTexColor,
	kUniformTexNormal,
	kUniformTexLUT,
	kUniformCount
};
static const char* kUniformNames[kUniformCount] = {
	"u_mvp",
	"u_world2object",
	"u_worldlightdir",
	"u_worldcampos",
	"u_lightcolor",
	"u_matcolor",
	"u_spec",
	"u_texcolor",
	"u_texnormal",
	"u_texLUT",
};
static GLint g_Uniforms[kUniformCount];

const int kGrid = 8;
const int kVertices = kGrid * kGrid;
const int kIndices = (kGrid-1)*(kGrid-1)*6;

static float kVBPosition[kVertices*3];
static float kVBNormal[kVertices*3];
static float kVBUV[kVertices*2];
static float kVBTan[kVertices*4];
static unsigned short kIB[kIndices];

static void InitializeMesh ()
{
	int idx = 0;
	for (int y = 0; y < kGrid; ++y)
	{
		for (int x = 0; x < kGrid; ++x, ++idx)
		{
			float u = float(x)/(kGrid-1);
			float v = float(y)/(kGrid-1);
			kVBPosition[idx*3+0] = u*2.0f-1.0f;
			kVBPosition[idx*3+1] = v*2.0f-1.0f;
			kVBPosition[idx*3+2] = 0.0f;
			kVBNormal[idx*3+0] = 0.0f;
			kVBNormal[idx*3+1] = 0.0f;
			kVBNormal[idx*3+2] = 1.0f;
			kVBUV[idx*2+0] = u*3.0f;
			kVBUV[idx*2+1] = v*3.0f;
			kVBTan[idx*4+0] = 1.0f;
			kVBTan[idx*4+1] = 0.0f;
			kVBTan[idx*4+2] = 0.0f;
			kVBTan[idx*4+3] = -1.0f;
		}
	}
	idx = 0;
	for (int y = 0; y < kGrid-1; ++y)
	{
		for (int x = 0; x < kGrid-1; ++x, idx+=6)
		{
			int base = y*kGrid+x;
			kIB[idx+0] = base;
			kIB[idx+1] = base+kGrid;
			kIB[idx+2] = base+1;
			kIB[idx+3] = base+kGrid;
			kIB[idx+4] = base+1;
			kIB[idx+5] = base+kGrid+1;
		}
	}
}


static void SetupMatrices (const float model[16], const float cam[16], const float proj[16])
{
	float view[16], modelview[16], mvp[16], invmodel[16];
	mat4f_Invert3D (cam, view);
	mat4f_Invert3D (model, invmodel);
	mat4f_MultiplyMat4f (view, model, modelview);
	mat4f_MultiplyMat4f (proj, modelview, mvp);
	glUniformMatrix4fv (g_Uniforms[kUniformMatrixMVP], 1, GL_FALSE, mvp);
	glUniformMatrix4fv (g_Uniforms[kUniformMatrixWorld2Object], 1, GL_FALSE, invmodel);
}




@interface ES2Renderer (PrivateMethods)
- (BOOL) loadShaders;
@end

@implementation ES2Renderer

// Create an ES 2.0 context
- (id <ESRenderer>) init
{
	if (self = [super init])
	{
		context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES2];
        
        if (!context || ![EAGLContext setCurrentContext:context] || ![self loadShaders])
		{
            [self release];
            return nil;
        }
		
		// Create default framebuffer object. The backing will be allocated for the current layer in -resizeFromLayer
		glGenFramebuffers(1, &defaultFramebuffer);
		glGenRenderbuffers(1, &colorRenderbuffer);
		glBindFramebuffer(GL_FRAMEBUFFER, defaultFramebuffer);
		glBindRenderbuffer(GL_RENDERBUFFER, colorRenderbuffer);
		glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_RENDERBUFFER, colorRenderbuffer);
		
		InitializeMesh ();
	}
	
	return self;
}

- (void)render
{
	static bool s_ProfilerInitialized = false;
	if (!s_ProfilerInitialized)
	{
		mach_timebase_info(&s_TimeInfo);
		s_ProfilerInitialized = true;
	}
	static int s_FrameCount = 0;
	Prof_Int64 time0 = 0, time1 = 0;
	time0 = mach_absolute_time();
	
    [EAGLContext setCurrentContext:context];
    
    glBindFramebuffer(GL_FRAMEBUFFER, defaultFramebuffer);
    glViewport(0, 0, backingWidth, backingHeight);
    
    glClearColor(0.5f, 0.4f, 0.5f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);
	
	// use shader program
	glUseProgram(program.prog);
	
	// update attribute values
	glVertexAttribPointer (kAttrPosition, 3, GL_FLOAT, false, 0, kVBPosition);
	glEnableVertexAttribArray(kAttrPosition);
	glVertexAttribPointer (kAttrNormal, 3, GL_FLOAT, false, 0, kVBNormal);
	glEnableVertexAttribArray(kAttrNormal);
	glVertexAttribPointer (kAttrTangent, 4, GL_FLOAT, false, 0, kVBTan);
	glEnableVertexAttribArray(kAttrTangent);
	glVertexAttribPointer (kAttrUV, 2, GL_FLOAT, false, 0, kVBUV);
	glEnableVertexAttribArray(kAttrUV);
	
    glActiveTexture (GL_TEXTURE0);
	glBindTexture (GL_TEXTURE_2D, 1);
    glUniform1i (g_Uniforms[kUniformTexColor], 0);
    glActiveTexture (GL_TEXTURE1);
	glBindTexture (GL_TEXTURE_2D, 2);
    glUniform1i (g_Uniforms[kUniformTexNormal], 1);
    glActiveTexture (GL_TEXTURE2);
	glBindTexture (GL_TEXTURE_2D, 3);
    glUniform1i (g_Uniforms[kUniformTexLUT], 2);
	
	// setup matrices
	float rot1[16], rot2[16], model[16], proj[16], cam[16];
	mat4f_LoadPerspective (0.45f, 2.0f/3.0f, 0.01f, 100.0f, proj);
	mat4f_LoadZRotation (rotz, rot1);
	mat4f_LoadXRotation (-M_PI*0.5f, rot2);
	mat4f_MultiplyMat4f (rot2, rot1, model);
	
	mat4f_LoadXRotation (-M_PI*0.25f, cam);
	cam[13] = 1.6f;
	cam[14] = 2.0f;
	SetupMatrices (model, cam, proj);
	
	float lx = 0.3f, ly=-1.0f, lz=1.0f;
	float invlen = 1.0f / sqrt(lx*lx+ly*ly+lz*lz);
	lx *= invlen; ly *= invlen; lz *= invlen;
	
	glUniform4f (g_Uniforms[kUniformVecLightDir], -lx, -ly, -lz, 0.0f);
	glUniform4f (g_Uniforms[kUniformVecCamPos], cam[12], cam[13], cam[14], 1.0f);
	glUniform4f (g_Uniforms[kUniformVecLightColor], kLightR, kLightG, kLightB, 1.0f);
	glUniform4f (g_Uniforms[kUniformVecMatColor], 1.0f, 1.0f, 1.0f, 1.0f);
	glUniform1f (g_Uniforms[kUniformFloatSpec], kSpec);
	
	// draw
	glDrawElements(GL_TRIANGLES, kIndices, GL_UNSIGNED_SHORT, kIB);
	
	rotz += 0.3f * M_PI / 180.0f;
	//rotz = 0.5f;

	glFinish ();
    
    glBindRenderbuffer(GL_RENDERBUFFER, colorRenderbuffer);
	time1 = mach_absolute_time();
	Prof_Int64 deltaTime = time1-time0;
	
    [context presentRenderbuffer:GL_RENDERBUFFER];
	
	if (s_FrameCount == 30)
	{
		printf("frametime avg: %4.1fms\n", MachToMillisecondsDelta(s_TimeAcc/30));
		s_TimeAcc = 0;
		s_FrameCount = 0;
	}
	s_TimeAcc += deltaTime;
	++s_FrameCount;
}

static void LoadTexture (GLint texID, NSString* path)
{
    NSData *texData = [[NSData alloc] initWithContentsOfFile:path];
    UIImage *image = [[UIImage alloc] initWithData:texData];
    if (image == nil)
		return;
	
    GLuint width = CGImageGetWidth(image.CGImage);
    GLuint height = CGImageGetHeight(image.CGImage);
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    void *imageData = malloc( height * width * 4 );
    CGContextRef context = CGBitmapContextCreate( imageData, width, height, 8, 4 * width, colorSpace, kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big );
    CGColorSpaceRelease( colorSpace );
    CGContextClearRect( context, CGRectMake( 0, 0, width, height ) );
    CGContextTranslateCTM( context, 0, height - height );
    CGContextDrawImage( context, CGRectMake( 0, 0, width, height ), image.CGImage );
	
	glBindTexture (GL_TEXTURE_2D, texID);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, imageData);
	
    CGContextRelease(context);
	
    free(imageData);
    [image release];
    [texData release];
}

static void CreateLightingTexture (GLint texID, int width, int height, float lr, float lg, float lb, float spec)
{
	unsigned char* data = new unsigned char[width*height*4];
	int idx = 0;
	for (int y = 0; y < height; ++y)
	{
		for (int x = 0; x < width; ++x, idx+=4)
		{
			float vx = float(x) / width;
			float vy = float(y) / height;
			float nl = vx;
			float nh = vy;
			float s = powf (nh, spec);
			data[idx+0] = nl * lr * 255.0f;
			data[idx+1] = nl * lg * 255.0f;
			data[idx+2] = nl * lb * 255.0f;
			data[idx+3] = s * 255.0f;
		}
	}
	glBindTexture (GL_TEXTURE_2D, texID);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
	delete[] data;
}

- (BOOL)loadShaders {
	
	// create shader program
	if (!LoadShader ([[NSBundle mainBundle] pathForResource:@"shader" ofType:@"txt"], &program))
	{
		return NO;
	}
	
	// get uniform locations
	for (int i = 0; i < kUniformCount; ++i)
		g_Uniforms[i] = glGetUniformLocation(program.prog, kUniformNames[i]);
	
	// create textures
	LoadTexture (1, [[NSBundle mainBundle] pathForResource:@"texturecolor" ofType:@"png"]);
	LoadTexture (2, [[NSBundle mainBundle] pathForResource:@"texturenormal" ofType:@"png"]);
	CreateLightingTexture (3, kLUTX, kLUTY, kLightR, kLightG, kLightB, kSpec);
	
	return YES;
}

- (BOOL) resizeFromLayer:(CAEAGLLayer *)layer
{
	// Allocate color buffer backing based on the current layer size
    glBindRenderbuffer(GL_RENDERBUFFER, colorRenderbuffer);
    [context renderbufferStorage:GL_RENDERBUFFER fromDrawable:layer];
	glGetRenderbufferParameteriv(GL_RENDERBUFFER, GL_RENDERBUFFER_WIDTH, &backingWidth);
    glGetRenderbufferParameteriv(GL_RENDERBUFFER, GL_RENDERBUFFER_HEIGHT, &backingHeight);
	
    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
	{
        NSLog(@"Failed to make complete framebuffer object %x", glCheckFramebufferStatus(GL_FRAMEBUFFER));
        return NO;
    }
	
    return YES;
}

- (void) dealloc
{
	// tear down GL
	if (defaultFramebuffer)
	{
		glDeleteFramebuffers(1, &defaultFramebuffer);
		defaultFramebuffer = 0;
	}
	
	if (colorRenderbuffer)
	{
		glDeleteRenderbuffers(1, &colorRenderbuffer);
		colorRenderbuffer = 0;
	}
	
	// realease the shader program object
	DestroyShader (&program);
	
	// tear down context
	if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];
	
	[context release];
	context = nil;
	
	[super dealloc];
}

@end
